--
-- dp_bmem_wrapper.vhd: Wrapper for dual port generic memory composed from
--                      Block RAMs
-- declaration
-- Copyright (C) 2014 CESNET
-- Author(s): Jiri Matousek <xmatou06@stud.fit.vutbr.cz>
--
-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions
-- are met:
-- 1. Redistributions of source code must retain the above copyright
--    notice, this list of conditions and the following disclaimer.
-- 2. Redistributions in binary form must reproduce the above copyright
--    notice, this list of conditions and the following disclaimer in
--    the documentation and/or other materials provided with the
--    distribution.
-- 3. Neither the name of the Company nor the names of its contributors
--    may be used to endorse or promote products derived from this
--    software without specific prior written permission.
--
-- This software is provided ``as is'', and any express or implied
-- warranties, including, but not limited to, the implied warranties of
-- merchantability and fitness for a particular purpose are disclaimed.
-- In no event shall the company or contributors be liable for any
-- direct, indirect, incidental, special, exemplary, or consequential
-- damages (including, but not limited to, procurement of substitute
-- goods or services; loss of use, data, or profits; or business
-- interruption) however caused and on any theory of liability, whether
-- in contract, strict liability, or tort (including negligence or
-- otherwise) arising in any way out of the use of this software, even
-- if advised of the possibility of such damage.
--
-- $Id$
--
-- TODO:


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_arith.all;

use WORK.math_pack.all;
use WORK.bmem_func.all;


-- ----------------------------------------------------------------------------
--                        Entity declaration
-- ----------------------------------------------------------------------------

entity DP_BMEM is
   -- Capacity of 1, 2, 4 Block rams is 16384 bits
   -- Capacity of 9, 18, 36 Block rams is 18432 bits
   generic (
      DATA_WIDTH      : integer := 32;
      -- Item in memory needed, one item size is DATA_WIDTH
      ITEMS           : integer := 1024;
      -- Block Ram Type, only 1, 2, 4, 9, 18, 36 bits
      BRAM_TYPE       : integer := 4;
      -- What operation will be performed first when both WE and RE are
      -- active? Only for behavioral simulation purpose.
      -- WRITE_FIRST(default) | READ_FIRST | NO_CHANGE
      WRITE_MODE_A    : string := "WRITE_FIRST";
      WRITE_MODE_B    : string := "WRITE_FIRST";
      -- Output directly from BlockRams or throw register
      -- TRUE, FALSE, AUTO (when column count > 1 then true)
      OUTPUT_REG      : TOUTPUT_REG := auto;
       -- Allow the output data register to be reset
      RESET_DATA_PATH : boolean := true;
      -- Debug prints
      DEBUG           : boolean := false
   );
   port (
      -- Interface A
      RSTA     : in  std_logic := '0';                         -- CLKA sync reset
      CLKA     : in  std_logic;                                -- Clock A
      PIPE_ENA : in  std_logic;                                -- Pipe Enable
      REA      : in  std_logic;                                -- Read Enable
      WEA      : in  std_logic;                                -- Write Enable
      ADDRA    : in  std_logic_vector(LOG2(ITEMS)-1 downto 0); -- Address A
      DIA      : in  std_logic_vector(DATA_WIDTH-1 downto 0);  -- Data A In
      DOA_DV   : out std_logic;                                -- Data A Valid
      DOA      : out std_logic_vector(DATA_WIDTH-1 downto 0);  -- Data A Out

      -- Interface B
      RSTB     : in  std_logic := '0';                         -- CLKB sync reset
      CLKB     : in  std_logic;                                -- Clock B
      PIPE_ENB : in  std_logic;                                -- Pipe Enable
      REB      : in  std_logic;                                -- Read Enable
      WEB      : in  std_logic;                                -- Write Enable
      ADDRB    : in  std_logic_vector(LOG2(ITEMS)-1 downto 0); -- Address B
      DIB      : in  std_logic_vector(DATA_WIDTH-1 downto 0);  -- Data B In
      DOB_DV   : out std_logic;                                -- Data B Valid
      DOB      : out std_logic_vector(DATA_WIDTH-1 downto 0)   -- Data B Out
   );
end entity DP_BMEM;


-- ----------------------------------------------------------------------------
--                           Architecture declaration
-- ----------------------------------------------------------------------------

architecture behavioral of DP_BMEM is
begin

   dp_bmem_i : entity work.DP_BMEM_TOP
   generic map (
      DATA_WIDTH      => DATA_WIDTH,
      ITEMS           => ITEMS,
      BRAM_TYPE       => BRAM_TYPE,
      WRITE_MODE_A    => WRITE_MODE_A,
      WRITE_MODE_B    => WRITE_MODE_B,
      OUTPUT_REG      => OUTPUT_REG,
      RESET_DATA_PATH => RESET_DATA_PATH,
      DEBUG           => DEBUG
   )
   port map (
      -- Interface A
      RSTA     => RSTA,
      CLKA     => CLKA,
      PIPE_ENA => PIPE_ENA,
      REA      => REA,
      WEA      => WEA,
      ADDRA    => ADDRA,
      DIA      => DIA,
      DOA_DV   => DOA_DV,
      DOA      => DOA,

      -- Interface B
      RSTB     => RSTB,
      CLKB     => CLKB,
      PIPE_ENB => PIPE_ENB,
      REB      => REB,
      WEB      => WEB,
      ADDRB    => ADDRB,
      DIB      => DIB,
      DOB_DV   => DOB_DV,
      DOB      => DOB
   );

end architecture behavioral;
